@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2002-2007 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2002-2009 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -23,8 +23,10 @@
 #include "affile.h"
 #include "driver.h"
 #include "messages.h"
-#include "macros.h"
 #include "misc.h"
+#include "serialize.h"
+#include "gprocess.h"
+#include "stats.h"
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -35,12 +37,16 @@
 #include <time.h>
 #include <stdlib.h>
 
+
 static gboolean
-affile_open_file(gchar *name, int flags,
-	     int uid, int gid, int mode,
-	     int dir_uid, int dir_gid, int dir_mode,
-	     int create_dirs, int *fd)
+affile_open_file(gchar *name, gint flags,
+                 uid_t uid, gid_t gid, mode_t mode,
+                 uid_t dir_uid, gid_t dir_gid, mode_t dir_mode,
+                 gboolean create_dirs, gboolean privileged, gboolean is_pipe, gint *fd)
 {
+  cap_t saved_caps;
+  struct stat st;
+
   if (strstr(name, "../") || strstr(name, "/..")) 
     {
       msg_error("Spurious path, logfile not created",
@@ -49,51 +55,64 @@ affile_open_file(gchar *name, int flags,
       return FALSE;
     }
 
-  *fd = open(name, flags, mode);
-  if (create_dirs && *fd == -1 && errno == ENOENT) 
+  if (create_dirs && !create_containing_directory(name, dir_uid, dir_gid, dir_mode))
+    return FALSE;
+
+  saved_caps = g_process_cap_save();
+  if (privileged)
     {
-      /* directory does not exist */
-      char *p = name + 1;
-      
-      p = strchr(p, '/');
-      while (p) 
-	{
-	  struct stat st;
-	  *p = 0;
-	  if (stat(name, &st) == 0) 
-	    {
-	      if (!S_ISDIR(st.st_mode))
-		return FALSE;
-	    }
-	  else if (errno == ENOENT) 
-	    {
-	      if (mkdir(name, dir_mode) == -1)
-		return 0;
-	      if (dir_uid != -1 || dir_gid != -1)
-		chown(name, dir_uid, dir_gid);
-	      if (dir_mode != -1)
-		chmod(name, dir_mode);
-	    }
-	  *p = '/';
-	  p = strchr(p + 1, '/');
-	}
-      *fd = open(name, flags, mode);
+      g_process_cap_modify(CAP_DAC_READ_SEARCH, TRUE);
+      g_process_cap_modify(CAP_SYS_ADMIN, TRUE);
     }
+  if (stat(name, &st) >= 0)
+    {
+      if (is_pipe && !S_ISFIFO(st.st_mode))
+        {
+          msg_error("Error opening pipe, underlying file is not a FIFO, it should be used by file()",
+                    evt_tag_str("filename", name),
+                    NULL);
+          goto exit;
+        }
+      else if (!is_pipe && S_ISFIFO(st.st_mode))
+        {
+          msg_error("Error opening file, underlying file is a FIFO, it should be used by pipe()",
+                    evt_tag_str("filename", name),
+                    NULL);
+          goto exit;
+        }
+    }
+  *fd = open(name, flags, mode);
+  if (is_pipe && *fd < 0 && errno == ENOENT)
+    {
+      if (mkfifo(name, 0666) >= 0)
+        *fd = open(name, flags, 0666);
+    }
+
   if (*fd != -1)
     {
       g_fd_set_cloexec(*fd, TRUE);
-      if (uid != -1)
+      
+      g_process_cap_modify(CAP_CHOWN, TRUE);
+      g_process_cap_modify(CAP_FOWNER, TRUE);
+      if (uid != (uid_t) -1)
         fchown(*fd, uid, -1);
-      if (gid != -1)
+      if (gid != (gid_t) -1)
         fchown(*fd, -1, gid);
-      if (mode != -1)
+      if (mode != (mode_t) -1)
         fchmod(*fd, mode);
     }
+ exit:
+  g_process_cap_restore(saved_caps);
+  msg_trace("affile_open_file",
+            evt_tag_str("path", name),
+            evt_tag_int("fd",*fd),
+            NULL);
+
   return *fd != -1;
 }
 
 static gboolean
-affile_sd_open_file(AFFileSourceDriver *self, gint *fd)
+affile_sd_open_file(AFFileSourceDriver *self, gchar *name, gint *fd)
 {
   gint flags;
   
@@ -102,11 +121,9 @@ affile_sd_open_file(AFFileSourceDriver *self, gint *fd)
   else
     flags = O_RDONLY | O_NOCTTY | O_NONBLOCK | O_LARGEFILE;
 
-  if (affile_open_file(self->filename->str, flags, -1, -1, -1, 0, 0, 0, 0, fd))
+  if (affile_open_file(name, flags, -1, -1, -1, 0, 0, 0, 0, !!(self->flags & AFFILE_PRIVILEGED), !!(self->flags & AFFILE_PIPE), fd))
     return TRUE;
-    
   return FALSE;
-
 }
 
 static inline gchar *
@@ -117,30 +134,107 @@ affile_sd_format_persist_name(AFFileSourceDriver *self)
   g_snprintf(persist_name, sizeof(persist_name), "affile_sd_curpos(%s)", self->filename->str);
   return persist_name;
 }
+ 
+static void
+affile_sd_save_pos(LogPipe *s, GlobalConfig *cfg)
+{
+  AFFileSourceDriver *self = (AFFileSourceDriver *) s;
+
+  if (cfg->persist && (self->flags & AFFILE_PIPE) == 0)
+    {
+      SerializeArchive *archive;
+      GString *str = g_string_sized_new(0);
+  
+      archive = serialize_string_archive_new(str);
+      log_reader_save_state((LogReader *) self->reader, archive);
+      cfg_persist_config_add_survivor(cfg, affile_sd_format_persist_name(self), str->str,  str->len, TRUE);
+      serialize_archive_free(archive);
+      g_string_free(str, TRUE);
+    }
+}
+
+static void
+affile_sd_load_pos(LogPipe *s, GlobalConfig *cfg)
+{
+  AFFileSourceDriver *self = (AFFileSourceDriver *) s;
+  gchar *str;
+  off_t cur_pos;
+  gsize str_len;
+  gint version;
+  
+  if ((self->flags & AFFILE_PIPE) || !cfg->persist)
+    return;
+     
+  str = cfg_persist_config_fetch(cfg, affile_sd_format_persist_name(self), &str_len, &version);
+  if (!str)
+    return;
+
+  if (version == 2)
+    {
+      /* NOTE: legacy, should be removed once the release after 3.0 is published */
+      cur_pos = strtoll(str, NULL, 10);
+      log_reader_update_pos((LogReader *) self->reader, cur_pos);
+      g_free(str);
+    }
+  else if (version >= 3)
+    {
+      GString *g_str = g_string_new("");
+      SerializeArchive *archive;
+
+      g_str = g_string_assign_len(g_str, str, str_len);
+      archive = serialize_string_archive_new(g_str);
+      
+      log_reader_restore_state((LogReader *) self->reader, archive);
+
+      serialize_archive_free(archive);
+      g_string_free(g_str, TRUE);
+      g_free(str);
+      
+    }
+}
 
 static void
 affile_sd_notify(LogPipe *s, LogPipe *sender, gint notify_code, gpointer user_data)
 {
   AFFileSourceDriver *self = (AFFileSourceDriver *) s;
-
+  GlobalConfig *cfg = log_pipe_get_config(s);
+  gint fd;
+  
   switch (notify_code)
     {
     case NC_FILE_MOVED:
       { 
-        gint fd;
-        
         msg_verbose("Follow-mode file source moved, tracking of the new file is started",
                     evt_tag_str("filename", self->filename->str),
                     NULL);
         
-        log_pipe_deinit(self->reader, NULL, NULL);
+        log_pipe_deinit(self->reader);
         log_pipe_unref(self->reader);
         
-        if (affile_sd_open_file(self, &fd))
+        if (affile_sd_open_file(self, self->filename->str, &fd))
           {
-            self->reader = log_reader_new(fd_read_new(fd, 0), LR_LOCAL | (self->reader_options.follow_freq == 0 ? LR_NOMREAD : 0), s, &self->reader_options);
+            LogTransport *transport;
+            
+            transport = log_transport_plain_new(fd, 0);
+            transport->timeout = 10;
+
+            self->reader = log_reader_new(
+                                  log_proto_plain_new_server(transport, self->reader_options.padding,
+                                                             self->reader_options.msg_size,
+                                                             ((self->reader_options.follow_freq > 0)
+                                                                    ? LPPF_IGNORE_EOF
+                                                                    : LPPF_NOMREAD)
+                                                            ),
+                                  LR_LOCAL);
+
+            log_reader_set_options(self->reader, s, &self->reader_options, 1, SCS_FILE, self->super.id, self->filename->str);
+
+            log_reader_set_follow_filename(self->reader, self->filename->str);
+      
+            log_reader_set_immediate_check(self->reader);
+
             log_pipe_append(self->reader, s);
-            if (!log_pipe_init(self->reader, NULL, NULL))
+            if (!log_pipe_init(self->reader, cfg))
               {
                 msg_error("Error initializing log_reader, closing fd",
                           evt_tag_int("fd", fd),
@@ -156,59 +250,69 @@ affile_sd_notify(LogPipe *s, LogPipe *sender, gint notify_code, gpointer user_da
           }
         break;
       }
+    default:
+      break;
     }
 }
 
-static gboolean
-affile_sd_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
+static void
+affile_sd_queue(LogPipe *s, LogMessage *msg, const LogPathOptions *path_options)
 {
   AFFileSourceDriver *self = (AFFileSourceDriver *) s;
+  
+  log_msg_add_dyn_value(msg, "FILE_NAME", self->filename->str);
+
+  log_pipe_forward_msg(s, msg, path_options);
+}
+
+static gboolean
+affile_sd_init(LogPipe *s)
+{
+  AFFileSourceDriver *self = (AFFileSourceDriver *) s;
+  GlobalConfig *cfg = log_pipe_get_config(s);
   gint fd;
   gboolean file_opened, open_deferred = FALSE;
 
-  log_reader_options_init(&self->reader_options, cfg);
+  log_reader_options_init(&self->reader_options, cfg, self->super.group);
 
-  file_opened = affile_sd_open_file(self, &fd);
+  file_opened = affile_sd_open_file(self, self->filename->str, &fd);
   if (!file_opened && self->reader_options.follow_freq > 0)
     {
       msg_info("Follow-mode file source not found, deferring open",
-	       evt_tag_str("filename", self->filename->str),
-	       NULL);
+               evt_tag_str("filename", self->filename->str),
+               NULL);
       open_deferred = TRUE;
       fd = -1;
     }
 
   if (file_opened || open_deferred)
     {
-      FDRead *fdr;
-      
-      fdr = fd_read_new(fd, 0);
-      fdr->timeout = 10;
-      self->reader = log_reader_new(fdr, LR_LOCAL | (self->reader_options.follow_freq == 0 ? LR_NOMREAD : 0), s, &self->reader_options);
+      LogTransport *transport;
 
-      if (persist)
-        {
-          gchar *str;
-          off_t cur_pos;
-          
-          str = persist_config_fetch(persist, affile_sd_format_persist_name(self));
-          if (str)
-            {
-              /* if the file could not be opened, we ignore the last remembered
-               * file position, if the file is created in the future we're going
-               * to read from the start. */
-              
-              if (file_opened)
-                {
-                  cur_pos = strtoll(str, NULL, 10);
-                  log_reader_set_pos((LogReader *) self->reader, cur_pos);
-                }
-              g_free(str);
-            }
-          
-        }
+      transport = log_transport_plain_new(fd, 0);
+      transport->timeout = 10;
+
+      /* FIXME: we shouldn't use reader_options to store log protocol parameters */
+      self->reader = log_reader_new(
+                            log_proto_plain_new_server(transport, self->reader_options.padding,
+                                                       self->reader_options.msg_size,
+                                                       ((self->reader_options.follow_freq > 0)
+                                                            ? LPPF_IGNORE_EOF
+                                                            : LPPF_NOMREAD)
+                                                      ),
+                            LR_LOCAL);
+      log_reader_set_options(self->reader, s, &self->reader_options, 1, SCS_FILE, self->super.id, self->filename->str);
+
+      log_reader_set_follow_filename(self->reader, self->filename->str);
+
+      /* NOTE: if the file could not be opened, we ignore the last
+       * remembered file position, if the file is created in the future
+       * we're going to read from the start. */
+      
+      affile_sd_load_pos(s, cfg);
       log_pipe_append(self->reader, s);
-      if (!log_pipe_init(self->reader, NULL, NULL))
+
+      if (!log_pipe_init(self->reader, NULL))
         {
           msg_error("Error initializing log_reader, closing fd",
                     evt_tag_int("fd", fd),
@@ -228,29 +332,23 @@ affile_sd_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
       return self->super.optional;
     }
   return TRUE;
+
 }
 
 static gboolean
-affile_sd_deinit(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
+affile_sd_deinit(LogPipe *s)
 {
   AFFileSourceDriver *self = (AFFileSourceDriver *) s;
+  GlobalConfig *cfg = log_pipe_get_config(s);
 
   if (self->reader)
     {
-      if (persist)
-        {
-          gchar str[32];
-          off_t cur_pos;
-          
-          cur_pos = log_reader_get_pos((LogReader *) self->reader);
-          g_snprintf(str, sizeof(str), "%" G_GINT64_FORMAT, (gint64) cur_pos);
-              
-          persist_config_add_survivor(persist, affile_sd_format_persist_name(self), str);
-        }
-      log_pipe_deinit(self->reader, NULL, NULL);
+      affile_sd_save_pos(s, cfg);
+      log_pipe_deinit(self->reader);
       log_pipe_unref(self->reader);
       self->reader = NULL;
     }
+
   return TRUE;
 }
 
@@ -262,28 +360,58 @@ affile_sd_free(LogPipe *s)
   g_string_free(self->filename, TRUE);
   g_assert(!self->reader);
 
-  log_drv_free_instance(&self->super);
-  g_free(s);
+  log_reader_options_destroy(&self->reader_options);
+
+  log_drv_free(s);
 }
 
 LogDriver *
 affile_sd_new(gchar *filename, guint32 flags)
 {
   AFFileSourceDriver *self = g_new0(AFFileSourceDriver, 1);
-
+  
   log_drv_init_instance(&self->super);
   self->filename = g_string_new(filename);
   self->flags = flags;
   self->super.super.init = affile_sd_init;
+  self->super.super.queue = affile_sd_queue;
   self->super.super.deinit = affile_sd_deinit;
   self->super.super.notify = affile_sd_notify;
   self->super.super.free_fn = affile_sd_free;
   log_reader_options_defaults(&self->reader_options);
-  self->reader_options.follow_filename = self->filename->str;
+  
+  if (configuration && configuration->version < 0x0300)
+    {
+      static gboolean warned = FALSE;
+      
+      if (!warned)
+        {
+          msg_warning("WARNING: file source: default value of follow_freq in file sources is changing in 3.0 to '1' for all files except /proc/kmsg",
+                      NULL);
+          warned = TRUE;
+        }
+    }
+  else
+    {
+      if ((self->flags & AFFILE_PIPE) == 0)
+        self->reader_options.follow_freq = 1000;
+        
+#if __linux__
+      if (strcmp(filename, "/proc/kmsg") == 0)
+        {
+          self->reader_options.follow_freq = 0;
+        }
+#endif
+    }
+#if __linux__
+  if (strcmp(filename, "/proc/kmsg") == 0)
+    {
+      self->flags |= AFFILE_PRIVILEGED;
+    }
+#endif
   return &self->super;
 }
 
-
 struct _AFFileDestWriter
 {
   LogPipe super;
@@ -302,17 +430,18 @@ affile_dw_reapable(AFFileDestWriter *self)
 }
 
 static gboolean
-affile_dw_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
+affile_dw_init(LogPipe *s)
 {
   AFFileDestWriter *self = (AFFileDestWriter *) s;
   int fd, flags;
   struct stat st;
+  GlobalConfig *cfg = log_pipe_get_config(s);
 
   if (cfg)
     self->time_reopen = cfg->time_reopen;
 
   msg_verbose("Initializing destination file writer",
-              evt_tag_str("template", self->owner->filename_template->template->str),
+              evt_tag_str("template", self->owner->filename_template->template),
               evt_tag_str("filename", self->filename->str),
               NULL);
               
@@ -336,17 +465,17 @@ affile_dw_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
   if (affile_open_file(self->filename->str, flags, 
                        self->owner->file_uid, self->owner->file_gid, self->owner->file_perm, 
                        self->owner->dir_uid, self->owner->dir_gid, self->owner->dir_perm, 
-                       !!(self->owner->flags & AFFILE_CREATE_DIRS), &fd))
+                       !!(self->owner->flags & AFFILE_CREATE_DIRS), FALSE, !!(self->owner->flags & AFFILE_PIPE), &fd))
     {
-      FDWrite *fdw;
+      guint write_flags;
       
       if (!self->writer)
         {
-          self->writer = log_writer_new(LW_FORMAT_FILE | ((self->owner->flags & AFFILE_PIPE) ? 0 : LW_ALWAYS_WRITABLE), s, &self->owner->writer_options);
-        
+          self->writer = log_writer_new(LW_FORMAT_FILE | ((self->owner->flags & AFFILE_PIPE) ? 0 : LW_ALWAYS_WRITABLE));
+          log_writer_set_options((LogWriter *) self->writer, s, &self->owner->writer_options, 1, self->owner->flags & AFFILE_PIPE ? SCS_PIPE : SCS_FILE, self->owner->super.id, self->filename->str);
           log_pipe_append(&self->super, self->writer);
         }
-      if (!log_pipe_init(self->writer, NULL, NULL))
+      if (!log_pipe_init(self->writer, NULL))
         {
           msg_error("Error initializing log writer", NULL);
           log_pipe_unref(self->writer);
@@ -354,11 +483,8 @@ affile_dw_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
           close(fd);
           return FALSE;
         }
-      fdw = fd_write_new(fd);
-      if (self->owner->flags & AFFILE_FSYNC)
-        fdw->fsync = TRUE;
-      fdw->append = TRUE;
-      log_writer_reopen(self->writer, fdw);
+      write_flags = ((self->owner->flags & AFFILE_FSYNC) ? LTF_FSYNC : 0) | LTF_APPEND;
+      log_writer_reopen(self->writer, log_proto_plain_new_client(log_transport_plain_new(fd, write_flags)));
     }
   else
     {
@@ -372,31 +498,32 @@ affile_dw_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
 }
 
 static gboolean
-affile_dw_deinit(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
+affile_dw_deinit(LogPipe *s)
 {
   AFFileDestWriter *self = (AFFileDestWriter *) s;
 
   if (self->writer)
     {
-      log_pipe_deinit(self->writer, NULL, NULL);
+      log_pipe_deinit(self->writer);
     }
   return TRUE;
 }
 
 static void
-affile_dw_queue(LogPipe *s, LogMessage *lm, gint path_flags)
+affile_dw_queue(LogPipe *s, LogMessage *lm, const LogPathOptions *path_options)
 {
   AFFileDestWriter *self = (AFFileDestWriter *) s;
+
   self->last_msg_stamp = time(NULL);
   if (!s->pipe_next && self->last_open_stamp < self->last_msg_stamp - self->time_reopen)
     {
-      log_pipe_init(&self->super, NULL, NULL);
+      log_pipe_init(&self->super, NULL);
     }
     
   if (s->pipe_next)
-    log_pipe_forward_msg(s, lm, path_flags);
+    log_pipe_forward_msg(s, lm, path_options);
   else
-    log_msg_drop(lm, path_flags);
+    log_msg_drop(lm, path_options);
 }
 
 static void
@@ -407,7 +534,7 @@ affile_dw_set_owner(AFFileDestWriter *self, AFFileDestDriver *owner)
   log_pipe_ref(&owner->super.super);
   self->owner = owner;
   if (self->writer)
-    log_writer_set_options((LogWriter *) self->writer, &self->super, &owner->writer_options);
+    log_writer_set_options((LogWriter *) self->writer, &self->super, &owner->writer_options, 1, SCS_FILE, self->owner->super.id, self->filename->str);
   
 }
 
@@ -420,8 +547,7 @@ affile_dw_free(LogPipe *s)
   self->writer = NULL;
   g_string_free(self->filename, TRUE);
   log_pipe_unref(&self->owner->super.super);
-  
-  g_free(s);
+  log_pipe_free(s);
 }
 
 static AFFileDestWriter *
@@ -546,26 +672,23 @@ affile_dd_set_fsync(LogDriver *s, gboolean fsync)
     self->flags &= ~AFFILE_FSYNC;
 }
 
+void
+affile_dd_set_local_time_zone(LogDriver *s, const gchar *local_time_zone)
+{
+  AFFileDestDriver *self = (AFFileDestDriver *) s;
+
+  self->local_time_zone = g_strdup(local_time_zone);
+}
+
 static inline gchar *
 affile_dd_format_persist_name(AFFileDestDriver *self)
 {
   static gchar persist_name[1024];
 
-  g_snprintf(persist_name, sizeof(persist_name), "affile_dd_writers(%s)", self->filename_template->template->str);
+  g_snprintf(persist_name, sizeof(persist_name), "affile_dd_writers(%s)", self->filename_template->template);
   return persist_name;
 }
 
-static const gchar *
-affile_dd_format_stats_name(AFFileDestDriver *self, gboolean is_pipe)
-{
-  static gchar stats_name[256];
-
-  g_snprintf(stats_name, sizeof(stats_name),
-             "%s(%s)", is_pipe ? "pipe" : "file", self->filename_template->template->str);
-
-  return stats_name;
-}
-
 static time_t reap_now = 0;
 
 static gboolean
@@ -577,10 +700,10 @@ affile_dd_reap_writers(gpointer key, gpointer value, gpointer user_data)
   if ((reap_now - dw->last_msg_stamp) >= self->time_reap && affile_dw_reapable(dw))
     {
       msg_verbose("Destination timed out, reaping", 
-                  evt_tag_str("template", self->filename_template->template->str),
+                  evt_tag_str("template", self->filename_template->template),
                   evt_tag_str("filename", dw->filename->str),
                   NULL);
-      log_pipe_deinit(&dw->super, NULL, NULL);
+      log_pipe_deinit(&dw->super);
       log_pipe_unref(&dw->super);
       
       /* remove from hash table */
@@ -595,7 +718,7 @@ affile_dd_reap(gpointer s)
   AFFileDestDriver *self = (AFFileDestDriver *) s;
   
   msg_verbose("Reaping unused destination files",
-              evt_tag_str("template", self->filename_template->template->str),
+              evt_tag_str("template", self->filename_template->template),
               NULL);  
   reap_now = time(NULL);
   if (self->writer_hash)
@@ -619,15 +742,16 @@ affile_dd_reuse_writer(gpointer key, gpointer value, gpointer user_data)
   AFFileDestWriter *writer = (AFFileDestWriter *) value;
   
   affile_dw_set_owner(writer, self);
-  log_pipe_init(&writer->super, NULL, NULL);
+  log_pipe_init(&writer->super, NULL);
 }
 
 
 static gboolean
-affile_dd_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
+affile_dd_init(LogPipe *s)
 {
   AFFileDestDriver *self = (AFFileDestDriver *) s;
-  
+  GlobalConfig *cfg = log_pipe_get_config(s);
+
   if (cfg->create_dirs)
     self->flags |= AFFILE_CREATE_DIRS;
   if (self->file_uid == -1)
@@ -646,28 +770,28 @@ affile_dd_init(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
     self->time_reap = cfg->time_reap;
   
   self->use_time_recvd = cfg->use_time_recvd;
+
+  if (self->local_time_zone_info)
+    time_zone_info_free(self->local_time_zone_info);
+  self->local_time_zone_info = time_zone_info_new(self->local_time_zone);
   
-  if (self->flags & AFFILE_PIPE)
-    log_writer_options_init(&self->writer_options, cfg, LWOF_SHARE_STATS, affile_dd_format_stats_name(self, TRUE));
-  else    
-    log_writer_options_init(&self->writer_options, cfg, LWOF_NO_STATS, NULL);
-    
-  self->cfg = cfg;
+  log_writer_options_init(&self->writer_options, cfg, 0);
+              
   
   if ((self->flags & AFFILE_NO_EXPAND) == 0)
     {
       self->reap_timer = g_timeout_add_full(G_PRIORITY_DEFAULT, self->time_reap * 1000 / 2, affile_dd_reap, self, NULL);
-      self->writer_hash = persist_config_fetch(persist, affile_dd_format_persist_name(self));
+      self->writer_hash = cfg_persist_config_fetch(cfg, affile_dd_format_persist_name(self), NULL, NULL);
       if (self->writer_hash)
         g_hash_table_foreach(self->writer_hash, affile_dd_reuse_writer, self);
     }
   else
     {
-      self->single_writer = persist_config_fetch(persist, affile_dd_format_persist_name(self));
+      self->single_writer = cfg_persist_config_fetch(cfg, affile_dd_format_persist_name(self), NULL, NULL);
       if (self->single_writer)
         {
           affile_dw_set_owner(self->single_writer, self);
-          log_pipe_init(&self->single_writer->super, NULL, NULL);
+          log_pipe_init(&self->single_writer->super, cfg);
         }
     }
   
@@ -684,7 +808,7 @@ static void
 affile_dd_destroy_writer(gpointer value)
 {
   AFFileDestWriter *writer = (AFFileDestWriter *) value;
-  log_pipe_deinit(&writer->super, NULL, NULL);
+  log_pipe_deinit(&writer->super);
   log_pipe_unref(&writer->super);
 }
 
@@ -717,22 +841,22 @@ affile_dd_destroy_writer_hash(gpointer value)
 static void
 affile_dd_deinit_writer(gpointer key, gpointer value, gpointer user_data)
 {
-  log_pipe_deinit((LogPipe *) value, NULL, NULL);
+  log_pipe_deinit((LogPipe *) value);
 }
 
 static gboolean
-affile_dd_deinit(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
+affile_dd_deinit(LogPipe *s)
 {
   AFFileDestDriver *self = (AFFileDestDriver *) s;
-
+  GlobalConfig *cfg = log_pipe_get_config(s);
   /* NOTE: we free all AFFileDestWriter instances here as otherwise we'd
    * have circular references between AFFileDestDriver and file writers */
   if (self->single_writer)
     {
       g_assert(self->writer_hash == NULL);
 
-      log_pipe_deinit(&self->single_writer->super, NULL, NULL);
-      persist_config_add(persist, affile_dd_format_persist_name(self), self->single_writer, affile_dd_destroy_writer);
+      log_pipe_deinit(&self->single_writer->super);
+      cfg_persist_config_add(cfg, affile_dd_format_persist_name(self), self->single_writer, -1, affile_dd_destroy_writer, FALSE);
       self->single_writer = NULL;
     }
   else if (self->writer_hash)
@@ -740,27 +864,27 @@ affile_dd_deinit(LogPipe *s, GlobalConfig *cfg, PersistentConfig *persist)
       g_assert(self->single_writer == NULL);
       
       g_hash_table_foreach(self->writer_hash, affile_dd_deinit_writer, NULL);
-      persist_config_add(persist, affile_dd_format_persist_name(self), self->writer_hash, affile_dd_destroy_writer_hash);
+      cfg_persist_config_add(cfg, affile_dd_format_persist_name(self), self->writer_hash, -1, affile_dd_destroy_writer_hash, FALSE);
       self->writer_hash = NULL;
     }
   if (self->reap_timer)
     g_source_remove(self->reap_timer);
-  self->cfg = NULL;
   return TRUE;
 }
 
 static void
-affile_dd_queue(LogPipe *s, LogMessage *msg, gint path_flags)
+affile_dd_queue(LogPipe *s, LogMessage *msg, const LogPathOptions *path_options)
 {
   AFFileDestDriver *self = (AFFileDestDriver *) s;
   AFFileDestWriter *next;
+  GlobalConfig *cfg = log_pipe_get_config(s);
 
   if (self->flags & AFFILE_NO_EXPAND)
     {
       if (!self->single_writer)
 	{
-	  next = affile_dw_new(self, g_string_new(self->filename_template->template->str));
-	  if (next && log_pipe_init(&next->super, self->cfg, NULL))
+	  next = affile_dw_new(self, g_string_new(self->filename_template->template));
+          if (next && log_pipe_init(&next->super, cfg))
 	    {
 	      self->single_writer = next;
 	    }
@@ -784,17 +908,18 @@ affile_dd_queue(LogPipe *s, LogMessage *msg, gint path_flags)
 
       filename = g_string_sized_new(32);
       log_template_format(self->filename_template, msg, 
-		    ((self->flags & AFFILE_TMPL_ESCAPE) ? MF_ESCAPE_RESULT : 0) |
-		    (self->use_time_recvd ? MF_STAMP_RECVD : 0),
-		    TS_FMT_BSD,
-		    get_local_timezone_ofs(time(NULL)),
-		    0,
-		    filename);
+                    ((self->flags & AFFILE_TMPL_ESCAPE) ? LT_ESCAPE : 0) |
+                    (self->use_time_recvd ? LT_STAMP_RECVD : 0),
+                    TS_FMT_BSD,
+                    self->local_time_zone_info,
+                    0,
+                    0,
+                    filename);
       next = g_hash_table_lookup(self->writer_hash, filename->str);
       if (!next)
 	{
 	  next = affile_dw_new(self, filename);
-	  if (!log_pipe_init(&next->super, self->cfg, NULL))
+          if (!log_pipe_init(&next->super, cfg))
 	    {
 	      log_pipe_unref(&next->super);
 	      next = NULL;
@@ -806,9 +931,9 @@ affile_dd_queue(LogPipe *s, LogMessage *msg, gint path_flags)
         g_string_free(filename, TRUE);
     }
   if (next)
-    log_pipe_queue(&next->super, msg, path_flags);
+    log_pipe_queue(&next->super, msg, path_options);
   else
-    log_msg_drop(msg, path_flags);
+    log_msg_drop(msg, path_options);
 }
 
 static void
@@ -821,8 +946,11 @@ affile_dd_free(LogPipe *s)
 
   log_template_unref(self->filename_template);
   log_writer_options_destroy(&self->writer_options);
-  log_drv_free_instance(&self->super);
-  g_free(self);
+  if (self->local_time_zone_info)
+    time_zone_info_free(self->local_time_zone_info);
+  if (self->local_time_zone)
+    g_free(self->local_time_zone);
+  log_drv_free(s);
 }
 
 LogDriver *
